import inspect
import typing as t
from attrs import define
import collections.abc as c
from contextlib import AsyncExitStack, AbstractAsyncContextManager

from .functional import to_context_manager, get_signature


type Cache = dict[t.Any, t.Any]
type Context = dict[str, t.Any]
type DepFn = t.Callable[..., t.Any]
type BaseFn[T] = t.Callable[..., c.Coroutine[t.Any, t.Any, T]]


__all__ = [
    "depend",
    "resolve",
    "Dependent",
    "get_dependant",
    "call_with_dependencies",
]


@define(slots=True, kw_only=True)
class Dependent:

    call: t.Callable[..., t.Any]


def depend(call: t.Callable[..., t.Any]) -> Dependent:
    return Dependent(call=call)


async def call_with_dependencies[T](fn: BaseFn[T], context: Context | None = None) -> T:
    async with AsyncExitStack() as stack:
        kwargs = await resolve(fn, cache={}, context=context or {}, exit_stack=stack, no_call=True)
        return await fn(**kwargs)


def get_dependant(annotation: t.Any) -> Dependent | None:
    if t.get_origin(annotation) is not t.Annotated:
        return None

    _, *metadata = t.get_args(annotation)
    for meta in metadata:
        if isinstance(meta, Dependent):
            return meta
    return None


async def resolve(
    fn: DepFn, cache: Cache, context: Context, exit_stack: AsyncExitStack, no_call: bool = False
) -> t.Any:
    if fn in cache:
        return cache[fn]

    kwargs: dict[str, t.Any] = {}

    for param in get_signature(fn):
        if context.get(param.name) is not None:
            kwargs[param.name] = context[param.name]
            continue

        dep = get_dependant(param.annotation)
        if dep is not None:
            kwargs[param.name] = await resolve(dep.call, cache=cache, context=context, exit_stack=exit_stack)
            continue

        if param.default is inspect.Parameter.empty:
            raise TypeError(f"Unresolvable dependency parameter: {param.name}")

        kwargs[param.name] = param.default

    if no_call:
        return kwargs

    result: t.Any = fn(**kwargs)
    result = to_context_manager(result) if inspect.isasyncgen(result) else result

    if isinstance(result, AbstractAsyncContextManager):
        value = await exit_stack.enter_async_context(result)
    elif inspect.isawaitable(result):
        value = await result
    else:
        value = result

    cache[fn] = value
    return value



import typing as t
from skoll.domain import Message
from skoll.result import Result, ok
from skoll.config import NATSConfig
from skoll.application.types import RawMessage
from skoll.application import Mediator, Subscriber, Service


__all__ = ["NATSMediator"]


class NATSMediator(Mediator):
    """NATS mediator implementation."""

    seed: str
    urls: list[str]
    subscribers: list[Subscriber]
    handlers: dict[str, Subscriber]

    def __init__(self, config: NATSConfig) -> None:
        if config.urls is None or config.seed is None:
            raise ValueError("NATS URLs and seed are required")
        self.urls = config.urls.split(",")
        self.seed = config.seed
        self.handlers = {}
        self.subscribers = []

    @t.override
    def register(self, *services: Service) -> None:
        for service in services:
            for subscriber in service.subscribers:
                if subscriber.with_reply:
                    self.handlers[subscriber.topic] = subscriber
                else:
                    self.subscribers.append(subscriber)

    @t.override
    async def connect(self) -> None:
        pass

    @t.override
    async def disconnect(self) -> None:
        pass

    @t.override
    async def publish(self, msg: Message | RawMessage) -> None:
        pass

    @t.override
    async def request(self, msg: Message | RawMessage) -> Result[t.Any]:
        return ok(None)



from __future__ import annotations

import re
import typing as t
from attrs import define, field
from datetime import datetime, timedelta, UTC

from skoll.errors import InvalidField
from skoll.domain.base import Object
from skoll.result import Result, ok, fail
from skoll.utils import new_ulid, to_tz, to_snake_case, safe_call

ID_REGEX = r"^[0-9a-z]{26}$"
EMAIL_REGEX = r"^[^@]+@[^@]+$"
TIME_REGEX = r"^(?:[01]?[0-9]|2[0-3]):[0-5][0-9]$"
LOCALE_PATTERN = r"^[a-z]{2,3}(-[A-Z][a-z]{3})?(-[A-Z]{2}|-[0-9]{3})?$"

__all__ = ["ID", "Time", "Email", "PositiveInt", "DateTime", "Latitude", "Longitude", "Locale", "LocalizedText"]


@define(kw_only=True, slots=True, frozen=True)
class ID(Object):

    value: str

    @classmethod
    def new(cls) -> t.Self:
        return cls(value=new_ulid())

    @t.override
    @classmethod
    def prepare(cls, raw: t.Any) -> Result[t.Any]:
        value = safe_call(str, raw)
        if value is not None and re.fullmatch(ID_REGEX, value.strip()) is not None:
            return ok(value.strip())
        return fail(
            InvalidField(
                field=to_snake_case(cls.__name__),
                hints={"expected": "string", "contraints": {"pattern": ID_REGEX}, "received": raw},
            )
        )


@define(kw_only=True, slots=True, frozen=True)
class PositiveInt(Object):

    value: int

    def increment(self) -> t.Self:
        return self.__class__(value=self.value + 1)

    def decrement(self) -> t.Self:
        return self.__class__(value=self.value - 1)

    @classmethod
    def zero(cls) -> t.Self:
        return cls(value=0)

    @t.override
    @classmethod
    def prepare(cls, raw: t.Any) -> Result[t.Any]:
        value = safe_call(int, raw)
        if value is not None and value >= 0:
            return ok(value)
        return fail(
            InvalidField(
                field=to_snake_case(cls.__name__),
                hints={"expected": "integer", "contraints": {"min": 0}, "received": raw},
            )
        )


@define(kw_only=True, slots=True, frozen=True)
class DateTime(Object):

    value: datetime

    @property
    def timestamp(self) -> int:
        return int(self.value.timestamp() * 1000)

    @property
    def week_day(self) -> int:
        return self.value.weekday()

    @property
    def iso_format(self) -> str:
        """ """
        return self.value.strftime("%Y-%m-%dT%H:%M:%S") + "Z"

    @t.override
    def serialize(self):
        return self.timestamp

    def diff(self, other: t.Self) -> timedelta:
        if other.value > self.value:
            return other.value - self.value
        return self.value - other.value

    def plus(
        self, days: int = 0, hours: int = 0, minutes: int = 0, seconds: int = 0, delta: timedelta | None = None
    ) -> DateTime:
        new_date = (
            self.value + timedelta(days=days, hours=hours, minutes=minutes, seconds=seconds) + (delta or timedelta())
        )
        return DateTime(value=new_date)

    def minus(
        self, days: int = 0, hours: int = 0, minutes: int = 0, seconds: int = 0, delta: timedelta | None = None
    ) -> DateTime:
        new_date = (
            self.value - timedelta(days=days, hours=hours, minutes=minutes, seconds=seconds) - (delta or timedelta())
        )
        return DateTime(value=new_date)

    def __gt__(self, other: t.Self) -> bool:
        return self.value > other.value

    def __lt__(self, other: t.Self) -> bool:
        return self.value < other.value

    def __ge__(self, other: t.Self) -> bool:
        return self.value >= other.value

    def __le__(self, other: t.Self) -> bool:
        return self.value <= other.value

    def reset_second(self):
        date = self.value.replace(second=0, microsecond=0)
        return DateTime(value=date)

    def reset_part(self, hour: bool = False, minute: bool = False):
        date = self.value.replace(second=0, microsecond=0)

        if hour:
            date = date.replace(hour=0, minute=0)
        elif minute:
            date = date.replace(minute=0)
        return DateTime(value=date)

    def to_tz(self, tz_str: str = "UTC") -> DateTime:
        return DateTime.from_timestamp(int(self.value.timestamp() * 1000), tz_str=tz_str)

    @classmethod
    def today(cls, tz_str: str = "UTC") -> DateTime:
        tz = to_tz(tz_str)
        value = datetime.combine(datetime.now(tz=tz), datetime.min.time()).replace(tzinfo=tz)
        return cls.from_timestamp(int(value.timestamp() * 1000), tz_str=tz_str)

    @classmethod
    def tomorrow(cls, tz_str: str = "UTC") -> DateTime:
        tz = to_tz(tz_str)
        value = datetime.combine(datetime.now(tz=tz) + timedelta(days=1), datetime.min.time()).replace(tzinfo=tz)
        return cls.from_timestamp(int(value.timestamp() * 1000), tz_str=tz_str)

    @classmethod
    def now(cls, tz_str: str = "UTC") -> DateTime:
        tz = to_tz(tz_str)
        value = int(datetime.now(tz=tz).timestamp() * 1000)
        return cls.from_timestamp(value, tz_str=tz_str)

    @classmethod
    def from_timestamp(cls, timestamp: int, tz_str: str = "UTC") -> DateTime:
        tz = to_tz(tz_str)
        return DateTime(value=datetime.fromtimestamp(timestamp / 1000, tz=tz))

    @t.override
    @classmethod
    def prepare(cls, raw: t.Any) -> Result[t.Any]:
        value = safe_call(int, raw)
        if value is not None and value >= 0:
            return ok(datetime.fromtimestamp(value / 1000, tz=UTC))
        return fail(
            InvalidField(
                field=to_snake_case(cls.__name__),
                hints={"expected": "integer", "contraints": {"min": 0}, "received": raw},
            )
        )


@define(kw_only=True, slots=True, frozen=True)
class Latitude(Object):
    value: float

    @t.override
    @classmethod
    def prepare(cls, raw: t.Any) -> Result[t.Any]:
        value = safe_call(float, raw)
        if value is not None and -90 <= value <= 90:
            return ok(value)
        return fail(
            InvalidField(
                field=to_snake_case(cls.__name__),
                hints={"expected": "float", "contraints": {"min": -90, "max": 90}, "received": raw},
            )
        )


@define(kw_only=True, slots=True, frozen=True)
class Longitude(Object):
    value: float

    @t.override
    @classmethod
    def prepare(cls, raw: t.Any) -> Result[t.Any]:
        value = safe_call(float, raw)
        if value is not None and -180 <= value <= 180:
            return ok(value)
        return fail(
            InvalidField(
                field=to_snake_case(cls.__name__),
                hints={"expected": "float", "contraints": {"min": -180, "max": 180}, "received": raw},
            )
        )


@define(kw_only=True, slots=True, frozen=True)
class Time(Object):
    value: str

    @property
    def as_hour(self) -> float:
        return self.hour + (self.minute / 60)

    @property
    def hour(self) -> int:
        return int(self.value.split(":")[0])

    @property
    def minute(self) -> int:
        return int(self.value.split(":")[1])

    @t.override
    @classmethod
    def prepare(cls, raw: t.Any) -> Result[t.Any]:
        value = safe_call(str, raw)
        if value is not None and re.fullmatch(TIME_REGEX, value.strip()) is not None:
            return ok(value.strip())

        return fail(
            InvalidField(
                field=to_snake_case(cls.__name__),
                hints={"expected": "string", "contraints": {"pattern": TIME_REGEX}, "received": raw},
            )
        )


@define(kw_only=True, slots=True, frozen=True)
class Email(Object):

    value: str

    @property
    def name(self) -> str:
        return self.value.split("@")[0]

    @classmethod
    def anonymous(cls, id: ID) -> Email:
        return cls(value=f"{id.value}.no-reply@email.com")

    @t.override
    @classmethod
    def prepare(cls, raw: t.Any) -> Result[t.Any]:
        value = safe_call(str, raw)
        if value is not None and re.fullmatch(EMAIL_REGEX, value.strip()) is not None:
            return ok(value.strip().lower())
        return fail(
            InvalidField(
                field=to_snake_case(cls.__name__),
                hints={"expected": "string", "contraints": {"pattern": EMAIL_REGEX}, "received": raw},
            )
        )


@define(kw_only=True, slots=True, frozen=True)
class Locale(Object):

    value: str

    @t.override
    @classmethod
    def prepare(cls, raw: t.Any) -> Result[t.Any]:
        value = safe_call(str, raw)
        if value is not None and re.fullmatch(LOCALE_PATTERN, value.strip()) is not None:
            return ok(value.strip().lower())
        return fail(
            InvalidField(
                field=to_snake_case(cls.__name__),
                hints={"received": raw, "expected": "BCP47Locale"},
            )
        )


@define(kw_only=True, slots=True, frozen=True)
class LocalizedText(Object):

    value: dict[str, str] = field(factory=dict)

    @t.override
    @classmethod
    def prepare(cls, raw: t.Any) -> Result[t.Any]:
        value = safe_call(dict, raw)

        if value is not None:
            keys_valid = all(isinstance(k, str) and re.fullmatch(LOCALE_PATTERN, k) for k in value.keys())
            values_valid = all(isinstance(v, str) for v in value.values())
            if keys_valid and values_valid:
                return ok(t.cast(dict[str, str], value))

        return fail(
            InvalidField(
                field=to_snake_case(cls.__name__),
                hints={
                    "received": raw,
                    "expected": "Dictionaire<BCP47Locale, string>",
                    "example": {"en-US": "English", "en": "An example"},
                },
            )
        )



import typing as t
from json import loads
from attrs import define
from re import match as re_match
from aiohttp import ClientSession

from skoll.utils import sanitize_dict
from skoll.config import SpiceDBConfig
from skoll.errors import Forbidden, InternalError
from skoll.application import AuthzWriteChange, AuthzPrecondition, AuthzLookupResult, Authz


__all__ = ["SpiceDBAuthz"]


TUPLE_PATTERN = r"(?P<resource>.+?):(?P<resource_id>.*?)#(?P<relation>.*?)@(?P<subject>.*?):(?P<subject_id>.*)"


@define(frozen=True, kw_only=True)
class TupleObject:
    resource: str
    subject: str | None = None
    relation: str | None = None
    subject_id: str | None = None
    resource_id: str | None = None
    subject_relation: str | None = None


class SpiceDBAuthz(Authz):

    url: str
    token: str

    def __init__(self, config: SpiceDBConfig):
        if config.uri is None or config.token is None:
            raise InternalError(debug={"message": "Invalid SpiceDB config", "config": config})
        self.url = config.uri
        self.token = config.token

    def make_headers(self) -> dict[str, str]:
        return {"Authorization": f"Bearer {self.token}"}

    @t.override
    async def write(self, changes: list[AuthzWriteChange], preconditions: list[AuthzPrecondition] | None = None) -> str:
        uri = self.url + "/v1/relationships/write"
        data = {"updates": get_changes(changes), "optionalAuthzPreconditions": get_preconditions(preconditions or [])}
        async with ClientSession() as session:
            async with session.post(uri, json=sanitize_dict(data), ssl=False, headers=self.make_headers()) as response:
                if response.status != 200:
                    print(await response.text())
                    raise ValueError(f"Failed to write changes {changes} with preconditions {preconditions}")
                return (await response.json()).get("writtenAt", {}).get("token", "")

    @t.override
    async def lookup(
        self, filter: str, cxt: dict[str, t.Any] | None = None, limit: int | None = None, cursor: str | None = None
    ) -> AuthzLookupResult:
        tuple_obj = tuple_from(filter)
        if not tuple_obj:
            raise ValueError(f"Invalid tuple: {filter}")

        data: dict[str, t.Any] = {
            "context": cxt,
            "consistency": {"minimizeLatency": True},
            "optionalCursor": {"token": cursor} if cursor else None,
        }
        uri = self.url + "/v1/permissions"
        if tuple_obj.subject_id is not None:  # ResourceLookup
            uri += "/resources"
            data["optionalLimit"] = limit
            data["permission"] = tuple_obj.relation
            data["resourceObjectType"] = tuple_obj.resource
            data["subject"] = {
                "optionalRelation": tuple_obj.subject_relation,
                "object": {"objectType": tuple_obj.subject, "objectId": tuple_obj.subject_id},
            }
        else:  # SubjectLookup
            uri += "/subjects"
            data["optionalConcreteLimit"] = limit
            data["permission"] = tuple_obj.relation
            data["subjectObjectType"] = tuple_obj.subject
            data["optionalSubjectRelation"] = tuple_obj.subject_relation
            data["resource"] = {"objectType": tuple_obj.resource, "objectId": tuple_obj.resource_id}

        async with ClientSession() as session:
            async with session.post(uri, json=sanitize_dict(data), ssl=False, headers=self.make_headers()) as response:
                if response.status != 200:
                    raise Exception(
                        f"Failed to lookup  {filter} with context {cxt} and limit {limit} and cursor {cursor}. status code: {response.status}"
                    )
                lines = [loads(line) for line in (await response.text()).split("\n")[:-1]]
                id_key = "resourceObjectId" if tuple_obj.subject_id is not None else "subjectObjectId"
                return AuthzLookupResult(
                    uids=[line.get("result", {}).get(id_key, "") for line in lines],
                    cursor=lines[-1].get("result", {}).get("afterResultCursor", {}).get("token"),
                )

    @t.override
    async def check(self, tuple: str, cxt: dict[str, t.Any] | None = None) -> None:
        uri = self.url + "/v1/permissions/check"
        tuple_obj = tuple_from(tuple)
        if not tuple_obj:
            raise ValueError(f"Invalid tuple: {tuple}")

        data: dict[str, t.Any] = {
            "context": cxt,
            "permission": tuple_obj.relation,
            "consistency": {"minimizeLatency": True},
            "resource": {"objectType": tuple_obj.resource, "objectId": tuple_obj.resource_id},
            "subject": {
                "optionalRelation": tuple_obj.subject_relation,
                "object": {"objectType": tuple_obj.subject, "objectId": tuple_obj.subject_id},
            },
        }

        async with ClientSession() as session:
            async with session.post(uri, json=sanitize_dict(data), ssl=False, headers=self.make_headers()) as response:
                res = await response.json()
                if res.get("permissionship", "") != "PERMISSIONSHIP_HAS_PERMISSION":
                    raise Forbidden(debug={"tuple": tuple, "context": cxt})
                return None


def tuple_from(tuple_str: str) -> TupleObject | None:
    match = re_match(TUPLE_PATTERN, tuple_str)
    if match:
        data = match.groupdict()
        subject_id = data["subject_id"].split("#")
        data["subject_id"] = subject_id[0]
        data["subject_relation"] = subject_id[1] if len(subject_id) > 1 else None
        return TupleObject(
            resource=data["resource"],
            subject_relation=data["subject_relation"],
            subject=data["subject"] if len(data["subject"]) > 0 else None,
            relation=data["relation"] if len(data["relation"]) > 0 else None,
            subject_id=data["subject_id"] if len(data["subject_id"]) > 0 else None,
            resource_id=data["resource_id"] if len(data["resource_id"]) > 0 else None,
        )
    return None


def get_changes(changes: list[AuthzWriteChange]) -> list[dict[str, t.Any]]:
    updates = []
    for ch in changes:
        tuple_obj = tuple_from(ch[1])
        if not tuple_obj:
            raise ValueError(f"Invalid tuple: {ch[1]}")
        update: dict[str, t.Any] = {
            "operation": "OPERATION_DELETE" if ch[0] == "DELETE" else "OPERATION_TOUCH",
            "relationship": {
                "relation": tuple_obj.relation,
                "resource": {"objectType": tuple_obj.resource, "objectId": tuple_obj.resource_id},
                "subject": {
                    "optionalRelation": tuple_obj.subject_relation,
                    "object": {"objectType": tuple_obj.subject, "objectId": tuple_obj.subject_id},
                },
                "optionalExpiresAt": ch[3].iso_format if ch[3] else None,
                "optionalCaveat": {"caveatName": ch[2][0], "context": ch[2][1]} if ch[2] else None,
            },
        }
        updates.append(update)
    return updates


def get_preconditions(precondition: list[AuthzPrecondition]) -> list[dict[str, t.Any]]:
    return [{"operation": f"OPERATION_{pr[0]}", "filter": relation_filter_from(pr[1])} for pr in precondition]


def relation_filter_from(tuple: str) -> dict[str, t.Any]:
    tuple_obj = tuple_from(tuple)
    if not tuple_obj:
        raise ValueError(f"Invalid tuple: {tuple}")
    return {
        "resourceType": tuple_obj.resource,
        "optionalResourceId": tuple_obj.resource_id,
        "optionalResourceIdPrefix": tuple_obj.resource_id,
        "optionalRelation": tuple_obj.relation,
        "optionalSubjectFilter": {
            "subjectType": tuple_obj.subject,
            "optionalSubjectId": tuple_obj.subject_id,
            "optionalRelation": {"relation": tuple_obj.subject_relation},
        },
    }



import typing as t

from skoll.domain import Message
from skoll.result import Result, fail, is_fail
from skoll.utils import call_with_dependencies
from skoll.application import Subscriber, RawMessage
from skoll.errors import Error, InternalError, ValidationFailed


async def run_callback(subscriber: Subscriber, message: Message | RawMessage) -> Result[t.Any]:
    topic = message.topic if isinstance(message, Message) else message.get("topic")
    try:
        msg = get_message(subscriber.msg_cls, message)
        return await call_with_dependencies(subscriber.callback, {subscriber.msg_arg: msg})
    except Error as err:
        return fail(err=err)
    except Exception as exc:
        return fail(err=InternalError.from_exception(exc, extra={"subject": topic, "message": message}))


def get_message(cls: type[Message], message: Message | RawMessage) -> Message:
    if isinstance(message, Message):
        return message
    res = cls.create(raw=message)
    if is_fail(res):
        raise ValidationFailed(errors=res.err.errors)

    return res.value


def is_subscribed(subscriber: Subscriber, message: Message | RawMessage) -> bool:
    topic = message.topic if isinstance(message, Message) else message.get("topic")
    s = topic.split(".") if topic else []
    p = subscriber.topic.split(".") if subscriber.topic else []

    i = j = 0
    while i < len(s) and j < len(p):
        if j + 1 < len(p) and p[j + 1] == ">":
            # Next token is '>', so this token must match exactly
            if p[j] != "*" and p[j] != s[i]:
                return False
            j += 2
            i += 1
            continue

        if p[j] == ">":
            return True

        if p[j] == "*":
            i += 1
            j += 1
            continue

        if p[j] != s[i]:
            return False

        i += 1
        j += 1

    if i == len(s):
        # Check remaining pattern tokens
        while j < len(p):
            if p[j] == ">":
                return True
            if p[j] != "*":
                return False
            j += 1
        return True

    # Subject has leftover tokens
    return j < len(p) and p[j] == ">"



import attrs
import typing as t

from skoll.utils import sanitize_dict


__all__ = [
    "Error",
    "NotFound",
    "Conflict",
    "Forbidden",
    "BadRequest",
    "InvalidField",
    "MissingField",
    "InternalError",
    "Unauthenticated",
    "ValidationFailed",
]

type ErrorStatusCode = t.Literal[500, 400, 401, 403, 404, 409]


@attrs.define(kw_only=True, slots=True)
class Error(Exception):

    code: str
    field: str | None = None
    status: ErrorStatusCode | None = None
    detail: str = "An expected error occurred."
    errors: list["Error"] = attrs.field(factory=list)
    debug: dict[str, t.Any] = attrs.field(factory=dict)
    hints: dict[str, t.Any] = attrs.field(factory=dict)

    def serialize(
        self,
        public_view: bool = False,
        trace_id: str | None = None,
    ) -> dict[str, t.Any]:
        err_json: dict[str, t.Any] = {
            "code": self.code,
            "field": self.field,
            "hints": self.hints,
            "trace_id": trace_id,
            "detail": self.detail,
            "status": self.status,
            "debug": self.debug if public_view is False else None,
            "errors": [sub.serialize(public_view=public_view) for sub in self.errors],
        }
        return sanitize_dict(err_json)


@attrs.define(kw_only=True, slots=True)
class MissingField(Error):

    code: str = attrs.field(default="missing_field", init=False)
    hints: dict[str, t.Any] = attrs.field(factory=dict, init=False)
    debug: dict[str, t.Any] = attrs.field(factory=dict, init=False)
    status: ErrorStatusCode | None = attrs.field(default=None, init=False)
    detail: str = attrs.field(default="This field is required to process your request", init=False)


@attrs.define(kw_only=True, slots=True)
class InvalidField(Error):

    code: str = attrs.field(default="invalid_field", init=False)
    debug: dict[str, t.Any] = attrs.field(factory=dict, init=False)
    status: ErrorStatusCode | None = attrs.field(default=None, init=False)
    detail: str = attrs.field(default="This field has invalid data, see hints for more details", init=False)


@attrs.define(kw_only=True, slots=True)
class InternalError(Error):

    field: str | None = attrs.field(default=None, init=False)
    code: str = attrs.field(default="internal_error", init=False)
    hints: dict[str, t.Any] = attrs.field(factory=dict, init=False)
    status: ErrorStatusCode | None = attrs.field(default=500, init=False)
    detail: str = attrs.field(default="An expected error occurred. Please try again later.", init=False)

    @classmethod
    def from_exception(cls, exc: Exception, extra: dict[str, t.Any] | None = None) -> t.Self:
        return cls(debug={**(extra or {}), **{"message": str(exc)}})


@attrs.define(kw_only=True, slots=True)
class BadRequest(Error):

    field: str | None = attrs.field(default=None, init=False)
    code: str = attrs.field(default="bad_request", init=False)
    status: ErrorStatusCode | None = attrs.field(default=400, init=False)
    detail: str = attrs.field(default="Your request is invalid. Check hints for more details", init=False)


@attrs.define(kw_only=True, slots=True)
class ValidationFailed(Error):

    field: str | None = attrs.field(default=None, init=False)
    debug: dict[str, t.Any] = attrs.field(factory=dict, init=False)
    code: str = attrs.field(default="validation_failed", init=False)
    status: ErrorStatusCode | None = attrs.field(default=400, init=False)
    detail: str = attrs.field(
        default="Your request contains invalid or missing data. Check hints for more details", init=False
    )


@attrs.define(kw_only=True, slots=True)
class Unauthenticated(Error):

    field: str | None = attrs.field(default=None, init=False)
    code: str = attrs.field(default="unauthenticated", init=False)
    status: ErrorStatusCode | None = attrs.field(default=401, init=False)
    detail: str = attrs.field(default="This request requires is only allowed for authenticated users", init=False)


@attrs.define(kw_only=True, slots=True)
class Forbidden(Error):

    code: str = attrs.field(default="forbidden", init=False)
    field: str | None = attrs.field(default=None, init=False)
    status: ErrorStatusCode | None = attrs.field(default=403, init=False)
    detail: str = attrs.field(
        default="You do not have required privilege for this operation. See hints for more details", init=False
    )


@attrs.define(kw_only=True, slots=True)
class NotFound(Error):

    code: str = attrs.field(default="not_found", init=False)
    field: str | None = attrs.field(default=None, init=False)
    status: ErrorStatusCode | None = attrs.field(default=404, init=False)
    detail: str = attrs.field(default="There is no resource corresponding to your request", init=False)


@attrs.define(kw_only=True, slots=True)
class Conflict(Error):

    code: str = attrs.field(default="conflict", init=False)
    field: str | None = attrs.field(default=None, init=False)
    status: ErrorStatusCode | None = attrs.field(default=409, init=False)
    detail: str = "Can not perform this operation since it will put the system in an inconsistent state"



import typing as t
from json import dumps
from attrs import define
import collections.abc as c
from asyncpg.pool import Pool, PoolConnectionProxy
from asyncpg import Record, create_pool, UniqueViolationError


from skoll.utils import from_json
from skoll.domain import EntityState
from skoll.result import Result, is_fail
from skoll.errors import InternalError, NotFound, Conflict
from skoll.application import DB, Repository, Criteria, ListCriteria, ListPage


__all__ = ["PostgresDB", "PostgresRepo", "parse_pg_row"]


def parse_pg_row(row: t.Any, errors_hints: dict[str, t.Any] | None = None) -> dict[str, t.Any]:
    raw = {}
    if row is None:
        raise NotFound(hints=errors_hints or {})
    if not isinstance(row, Record):
        raise InternalError(debug={"row": row, "message": "Invalid row PG data", "errors_hints": errors_hints})
    for key, value in row.items():
        json_value = from_json(value)
        raw[key] = json_value if isinstance(value, str) and json_value is not None else value
    return raw


class PostgresDB(DB[PoolConnectionProxy]):

    dsn: str
    __pool: Pool | None

    def __init__(self, dsn: str = "PG_DB_DSN") -> None:
        self.dsn = dsn
        self.__pool = None

    @t.override
    async def connect(self) -> None:
        if self.__pool is None:
            try:
                self.__pool = await create_pool(dsn=self.dsn, min_size=1, max_size=10)
            except Exception as exc:
                raise InternalError.from_exception(exc)

    @t.override
    async def close(self) -> None:
        if self.__pool is not None:
            await self.__pool.close()
            self.__pool = None

    @t.override
    async def session(self) -> c.AsyncGenerator[PoolConnectionProxy]:
        if self.__pool is None:
            raise RuntimeError("Database pool is not initialized.")
        async with self.__pool.acquire() as conn:
            yield conn

    @t.override
    async def transaction(self) -> c.AsyncGenerator[PoolConnectionProxy]:
        if self.__pool is None:
            raise RuntimeError("Database pool is not initialized.")
        async with self.__pool.acquire() as conn:
            async with conn.transaction():
                yield conn


@define(kw_only=True, frozen=True, slots=True)
class PostgresRepo[T: EntityState](Repository[T]):

    table: str
    conn: PoolConnectionProxy
    restore_func: t.Callable[[dict[str, t.Any]], Result[T]]

    @t.override
    async def get(self, criteria: Criteria) -> T | None:
        try:
            qry, params = criteria.as_sql
            record = await self.conn.fetchrow(qry, *params)
            if not isinstance(record, Record):
                return None
            res = self.restore_func(parse_pg_row(record))
            if is_fail(res):
                raise ValueError("Entity Parsing failed")
            return res.value
        except Exception as exc:
            raise InternalError.from_exception(exc, extra={"criteria": criteria.as_sql})

    @t.override
    async def exists(self, criteria: Criteria) -> bool:
        try:
            qry, params = criteria.as_sql
            record = await self.conn.fetchrow(qry, *params)
            return record is not None
        except Exception as exc:
            raise InternalError.from_exception(exc, extra={"criteria": criteria.as_sql})

    @t.override
    async def delete(self, criteria: Criteria) -> None:
        try:
            qry, params = criteria.as_sql
            await self.conn.execute(qry.replace("SELECT *", "DELETE"), *params)
        except Exception as exc:
            raise InternalError.from_exception(exc, extra={"criteria": criteria.as_sql})

    @t.override
    async def list(self, criteria: ListCriteria) -> ListPage[T]:
        try:
            qry, params = criteria.as_sql
            rows = await self.conn.fetch(qry, *params)
            items: list[T] = []
            for row in rows:
                if not isinstance(row, Record):
                    raise ValueError("Invalid row type")
                res = self.restore_func(parse_pg_row(row))
                if is_fail(res):
                    raise ValueError("Entity Parsing failed")
                items.append(res.value)
            return ListPage(cursor="NOOP", items=items)  # TODO: Implement cursor
        except Exception as exc:
            raise InternalError.from_exception(exc, extra={"criteria": criteria.as_sql})

    @t.override
    async def save(self, state: T) -> None:
        try:
            raw = state.serialize()
            sql_stm, params = self.__prepare_insert(raw) if state.version.value == 0 else self.__prepare_update(raw)
            _ = await self.conn.execute(sql_stm, *params)
        except UniqueViolationError as exc:
            raise Conflict(debug={"raw": state.serialize(), "table": self.table})
        except Exception as exc:
            raise InternalError.from_exception(exc, extra={"raw": state.serialize(), "table": self.table})

    def __prepare_insert(self, raw: dict[str, t.Any]):
        params: list[t.Any] = []
        attrs: list[str] = []
        keys: list[str] = []
        for idx, kv in enumerate(raw.items()):
            attrs.append(kv[0])
            keys.append(f"${idx + 1}")
            params.append(dumps(kv[1]) if isinstance(kv[1], (dict, list)) else kv[1])
        sql_stm = f"INSERT INTO {self.table}({", ".join(attrs)}) VALUES({", ".join(keys)})"
        return sql_stm, params

    def __prepare_update(self, raw: dict[str, t.Any]):
        params = [raw["uid"], raw["version"] - 1]
        changes: list[str] = []
        for idx, kv in enumerate(raw.items()):
            changes.append(f"{kv[0]} = ${idx + 3}")
            params.append(dumps(kv[1]) if isinstance(kv[1], (dict, list)) else kv[1])
        sql_stm = f"UPDATE {self.table} SET {", ".join(changes)} WHERE uid = $1 AND version = $2"
        return sql_stm, params



from certifi import where
from attrs import define, field
from ssl import SSLContext, create_default_context

from skoll.utils import get_config_var


__all__ = ["SSLContextConfig", "SpiceDBConfig", "PostgresConfig", "NATSConfig"]


@define(frozen=True, kw_only=True)
class SSLContextConfig:

    default: SSLContext = field(factory=lambda: create_default_context(cafile=where()))


@define(frozen=True, kw_only=True)
class SpiceDBConfig:

    uri: str | None = field(factory=get_config_var(["SPICEDB_URI", "/run/secrets/spicedb_uri.txt"]))
    token: str | None = field(factory=get_config_var(["SPICEDB_TOKEN", "/run/secrets/spicedb_token.txt"]))


@define(frozen=True, kw_only=True)
class PostgresConfig:

    dsn: str | None = field(factory=get_config_var(["PG_DB_DSN", "/run/secrets/pg_db_dsn.txt"]))


@define(frozen=True, kw_only=True)
class NATSConfig:

    urls: str | None = field(factory=get_config_var(["NATS_URLS", "/run/secrets/nats_urls.txt"]))
    seed: str | None = field(factory=get_config_var(["NATS_SEED", "/run/secrets/nats_seed.txt"]))



from .base import Enum

__all__ = ["SortDirection", "BaseStatus"]


class SortDirection(Enum):

    ASCENDING = "ASC"
    DESCENDING = "DESC"


class BaseStatus(Enum):

    ACTIVE = "ACTIVE"
    INACTIVE = "INACTIVE"
    DELETED = "DELETED"
    ARCHIVED = "ARCHIVED"



import typing as t
from attrs import define

from .errors import Error, InvalidField


__all__ = ["Result", "ok", "fail", "is_ok", "is_fail", "combine"]


type Result[T] = _Ok[T] | _Fail
V = t.TypeVar("V", covariant=True)


@define(frozen=True, slots=True)
class _Ok(t.Generic[V]):

    value: V


@define(frozen=True, slots=True)
class _Fail:

    err: Error


def ok(value: V) -> Result[V]:
    return _Ok(value)


def fail(err: Error) -> Result[t.Any]:
    return _Fail(err)


def is_ok[T](res: Result[T]) -> t.TypeIs[_Ok[T]]:
    return isinstance(res, _Ok)


def is_fail[T](res: Result[T]) -> t.TypeIs[_Fail]:
    return isinstance(res, _Fail)


@t.overload
def combine(results: list[Result[t.Any]]) -> Result[list[t.Any]]: ...


@t.overload
def combine(results: dict[str, Result[t.Any]]) -> Result[dict[str, t.Any]]: ...


def combine(results: list[Result[t.Any]] | dict[str, Result[t.Any]]):
    if isinstance(results, dict):
        dict_values = {k: r.value for k, r in results.items() if isinstance(r, _Ok)}
        errs = [r.err for r in results.values() if isinstance(r, _Fail)]
        return _Ok(dict_values) if not errs else _Fail(InvalidField(errors=errs))

    list_values = [r.value for r in results if isinstance(r, _Ok)]
    errs = [r.err for r in results if isinstance(r, _Fail)]
    return _Ok(list_values) if not errs else _Fail(InvalidField(errors=errs))



from __future__ import annotations

import typing as t
from datetime import timedelta
from attrs import define, field

from skoll.domain.base import Object
from skoll.domain.primitives import ID, PositiveInt, DateTime, Latitude, Longitude, Time, LocalizedText


__all__ = [
    "Period",
    "Address",
    "Message",
    "TimeSlot",
    "Coordinate",
    "EntityState",
    "RegularHours",
    "SpecialHours",
    "WorkingHours",
    "MessageContext",
    "MessagePayload",
]


@define(kw_only=True, slots=True, frozen=True)
class Coordinate(Object):

    lat: Latitude
    lng: Longitude

    @classmethod
    def from_raw(cls, lat: float, lng: float):
        return Coordinate(lat=Latitude(value=lat), lng=Longitude(value=lng))


@define(kw_only=True, slots=True, frozen=True)
class Address(Object):

    city: str
    street: str
    region: str
    country: str
    postal_code: str
    coordinate: Coordinate


@define(kw_only=True, slots=True, frozen=True)
class Period(Object):

    end: DateTime
    start: DateTime

    @property
    def duration(self) -> timedelta:
        return self.end.diff(self.start)

    def days(self, tz_str: str = "UTC") -> list[DateTime]:
        dates: list[DateTime] = []
        date = self.start.to_tz(tz_str=tz_str).reset_part(hour=True, minute=True).reset_second()
        while date < self.end.to_tz(tz_str=tz_str):
            dates.append(date)
            date = date.plus(days=1)

        return dates


@define(kw_only=True, slots=True, frozen=True)
class TimeSlot(Object):

    end: Time
    start: Time

    def is_between(self, hour: int, minute: int):
        hours = hour + (minute / 60)
        return self.start.as_hour <= hours and hours <= self.end.as_hour


@define(kw_only=True, slots=True, frozen=True)
class RegularHours(Object):

    weekday: list[int] = field(factory=list)
    slots: list[TimeSlot] = field(factory=list)


@define(kw_only=True, slots=True, frozen=True)
class SpecialHours(Object):

    opened: bool
    date: DateTime
    name: LocalizedText
    slots: list[TimeSlot] = field(factory=list)


@define(kw_only=True, slots=True, frozen=True)
class WorkingHours(Object):

    timezone: str
    always_open: bool
    regular_hours: list[RegularHours] = field(factory=list)
    special_hours: list[SpecialHours] = field(factory=list)


@define(kw_only=True, slots=True, frozen=True, eq=False)
class EntityState(Object):

    id: ID = field(factory=ID.new)
    created_at: DateTime = field(factory=DateTime.now)
    updated_at: DateTime = field(factory=DateTime.now)
    version: PositiveInt = field(factory=PositiveInt.zero)

    @t.override
    def __eq__(self, other: t.Any) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return other.__hash__() == self.__hash__()

    @t.override
    def __ne__(self, other: t.Any) -> bool:
        return not self == other

    @t.override
    def __hash__(self) -> int:
        return hash(self.id.serialize())

    @t.override
    def evolve(self, **kwargs: t.Any) -> t.Self:
        if "updated_at" not in kwargs:
            kwargs["updated_at"] = DateTime.now()
        if "version" not in kwargs:
            kwargs["version"] = self.version.increment()
        return super().evolve(**kwargs)


@define(frozen=True, kw_only=True, slots=True)
class MessageContext(Object):

    span_id: ID = field(factory=ID.new)
    trace_id: ID = field(factory=ID.new)

    @classmethod
    def default(cls) -> t.Self:
        return cls()


@define(frozen=True, kw_only=True, slots=True)
class MessagePayload(Object):

    @classmethod
    def default(cls) -> t.Self:
        return cls()


@define(frozen=True, kw_only=True, slots=True, eq=False)
class Message(Object):

    topic: str
    source: str
    id: ID = field(factory=ID.new)
    created_at: DateTime = field(factory=DateTime.now)
    payload: MessagePayload = field(factory=MessagePayload.default)
    context: MessageContext = field(factory=MessageContext.default)

    @t.override
    def __eq__(self, other: t.Any) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return other.__hash__() == self.__hash__()

    @t.override
    def __ne__(self, other: t.Any) -> bool:
        return not self == other

    @t.override
    def __hash__(self) -> int:
        return hash(self.id.serialize())



import attrs
import typing as t
from abc import ABC
from types import UnionType
from enum import Enum as _Enum


from skoll.utils import to_snake_case, serialize
from skoll.errors import MissingField, InvalidField, Error
from skoll.result import Result, fail, ok, combine, is_fail, is_ok


__all__ = ["Enum", "Object"]


class Enum(_Enum):

    def serialize(self) -> str:
        return self.value

    @classmethod
    def options(cls) -> list[str]:
        return [option.value for option in list(cls)]

    @classmethod
    def create(cls, raw: t.Any) -> Result[t.Self]:
        if raw in cls.options():
            return ok(cls(raw))

        if raw is None:
            return fail(MissingField(field=to_snake_case(cls.__name__)))

        return fail(
            Error(
                code="unknown_option",
                field=to_snake_case(cls.__name__),
                hints={"expected": cls.options(), "received": raw},
                detail=f"Invalid input passed, check the hints for more details.",
            )
        )


@attrs.define(kw_only=True, frozen=True, slots=True)
class Object(ABC):

    def serialize(self) -> t.Any:
        data = attrs.asdict(self)
        if len(data) == 1 and data.get("value") is not None:
            return serialize(data["value"])
        return serialize(data)

    @classmethod
    def prepare(cls, raw: t.Any) -> Result[t.Any]:
        return ok(raw)

    @classmethod
    def _init(cls, value: t.Any) -> Result[t.Self]:
        if isinstance(value, dict):
            return ok(cls(**value))
        return ok(cls(**{"value": value}))

    @classmethod
    def create(cls, raw: t.Any) -> Result[t.Self]:
        if raw is None:
            return fail(MissingField(field=to_snake_case(cls.__name__)))

        prepare_result = cls.prepare(raw)
        if is_fail(prepare_result):
            return prepare_result

        schema = get_schema(cls)
        if schema is None:
            return cls._init(prepare_result.value)

        results: dict[str, Result[t.Any]] = {}
        for key, item in schema.items():
            results[key] = item.create(prepare_result.value)
        res = combine(results)
        return cls._init(**res.value) if is_ok(res) else res

    def evolve(self, **kwargs: t.Any) -> t.Self:
        return attrs.evolve(self, **kwargs)


@attrs.define(kw_only=True, slots=True, frozen=True)
class _SchemaItem:
    """
    Represents a single item in object schema.
    """

    key: str
    cls: t.Any
    is_list: bool = False
    optional: bool = False
    default: t.Any = attrs.NOTHING

    def create(self, raw: t.Any) -> Result[t.Any]:
        raw_item = raw.get(self.key) if isinstance(raw, dict) else raw
        if raw_item is None and self.default != attrs.NOTHING:
            if not hasattr(self.default, "factory"):
                return ok(value=self.default)
            if self.default.takes_self is False:
                return ok(value=self.default.factory())
            # Ignore case where takes_self is True, as we don't have the object instance here

        if self.optional is True and raw_item is None:
            return ok(value=None)
        if raw_item is None:
            return fail(MissingField(field=self.key))
        if self.is_list and isinstance(raw_item, list):
            results: list[Result[t.Any]] = []
            for idx, rw in enumerate(t.cast(list[t.Any], raw_item)):
                res = self._create(raw=rw, field=f"{self.key}[{idx}]")
                results.append(res)
            res = combine(results)
            if is_fail(res):
                res.err.field = self.key
            return res
        if self.is_list and not isinstance(raw_item, list):
            return fail(InvalidField(field=self.key))
        return self._create(raw=raw_item)

    def _create(self, raw: t.Any, field: str | None = None) -> Result[t.Any]:
        field = field or self.key
        if hasattr(self.cls, "create") and callable(getattr(self.cls, "create")):
            return self.cls.create(raw)

        type_checkers: dict[type[t.Any], t.Callable[[t.Any], bool]] = {
            bool: lambda x: x in ["True", "False", True, False],
            str: lambda x: isinstance(x, (str, float, int)),
            int: lambda x: isinstance(x, (int, float)),
            float: lambda x: isinstance(x, (float, int)),
        }
        check = type_checkers.get(self.cls)

        if check is None or check(raw) is True:
            return ok(raw if check is None else self.cls(raw))

        return fail(InvalidField(field=field, hints={"expected": self.cls.__name__, "received": raw}))


def get_schema(cls: type[t.Any]) -> dict[str, _SchemaItem] | None:
    if not attrs.has(cls) or (len(attrs.fields(cls)) == 1 and attrs.fields(cls)[0].name == "value"):
        return None

    schema: dict[str, _SchemaItem] = {}
    for key, attr in attrs.fields_dict(cls).items():
        is_list, optional, _cls = False, False, attr.type
        if t.get_origin(_cls) == UnionType and t.get_args(_cls)[1] == type(None):
            _cls = t.get_args(attr.type)[0]
            optional = True
        if t.get_origin(_cls) == list:
            is_list = True
            _cls = t.get_args(attr.type)[0]

        schema[key] = _SchemaItem(key=key, cls=_cls, is_list=is_list, optional=optional, default=attr.default)
    return schema



import typing as t
import collections.abc as c
from inspect import signature
from attrs import define, field
from abc import ABC, abstractmethod

from skoll.result import Result
from skoll.errors import NotFound
from skoll.domain import DateTime, ID, Message


__all__ = [
    "Service",
    "Criteria",
    "ListPage",
    "Subscriber",
    "RawMessage",
    "ListCriteria",
    "AuthzWriteChange",
    "AuthzLookupResult",
    "MissingSubscriber",
    "AuthzPrecondition",
    "AuthzWriteOperation",
    "AuthzPreconditionOperation",
]


type AuthzWriteOperation = t.Literal["SET", "DELETE"]
type AuthzPrecondition = tuple[AuthzPreconditionOperation, str]
type AuthzPreconditionOperation = t.Literal["MUST_MATCH", "MUST_NOT_MATCH"]
type SubscriberCallback = t.Callable[t.Concatenate[Message, ...], c.Coroutine[t.Any, t.Any, Result[t.Any]]]
type AuthzWriteChange = tuple[AuthzWriteOperation, str, tuple[str, dict[str, t.Any]] | None, DateTime | None]


class RawMessage(t.TypedDict):
    topic: str
    source: str
    id: t.NotRequired[str]
    created_at: t.NotRequired[int]
    payload: t.NotRequired[dict[str, t.Any]]
    context: t.NotRequired[dict[str, t.Any]]


class ListPage[T](t.NamedTuple):

    items: list[T]
    cursor: str | None = None


class AuthzLookupResult(t.NamedTuple):
    uids: list[str]
    cursor: str | None = None


class SQLCriteria(t.NamedTuple):
    query: str
    params: list[t.Any]


@define(frozen=True, kw_only=True, slots=True)
class Criteria(ABC):

    uid: ID | None = None

    @property
    @abstractmethod
    def as_sql(self) -> SQLCriteria:
        raise NotImplementedError("Subclasses must implement this method")


@define(frozen=True, kw_only=True, slots=True)
class ListCriteria(Criteria, ABC):

    cursor: str | None = None
    limit: int = field(default=100)


@define(frozen=True, kw_only=True, slots=True)
class Subscriber:

    topic: str
    msg_arg: str
    service_name: str
    msg_cls: type[Message]
    with_reply: bool = True
    callback: SubscriberCallback


@define(kw_only=True, slots=True)
class MissingSubscriber(NotFound):

    attr: str | None = field(default=None, init=False)
    code: str = field(default="missing_subscriber", init=False)
    detail: str = "No subscriber found for the given message subject"


@define(kw_only=True, slots=True, frozen=True)
class Service:

    name: str
    subscribers: list[Subscriber] = field(factory=list)

    def on(self, topic: str, with_reply: bool = True):
        def decorator(callback: SubscriberCallback):
            first_arg = list(signature(callback).parameters.values())[0]
            if not issubclass(first_arg.annotation, Message):
                raise TypeError(
                    f"@service.on only allow on function with first argument being a subclass of Message, got {first_arg.annotation}"
                )

            self.subscribers.append(
                Subscriber(
                    topic=topic,
                    callback=callback,
                    with_reply=with_reply,
                    msg_arg=first_arg.name,
                    service_name=self.name,
                    msg_cls=first_arg.annotation,
                )
            )
            return callback

        return decorator



import typing as t
import collections.abc as c

from skoll.result import Result
from skoll.application.types import *
from skoll.domain import EntityState, Message


__all__ = ["DB", "Mediator", "Authz", "Repository"]


class Mediator(t.Protocol):

    async def connect(self) -> None: ...
    async def disconnect(self) -> None: ...
    def register(self, *services: Service) -> None: ...
    async def publish(self, msg: Message | RawMessage) -> None: ...
    async def request(self, msg: Message | RawMessage) -> Result[t.Any]: ...


class Authz(t.Protocol):

    async def check(self, tuple: str, cxt: dict[str, t.Any] | None = None) -> None: ...
    async def write(
        self, changes: list[AuthzWriteChange], preconditions: list[AuthzPrecondition] | None = None
    ) -> str: ...
    async def lookup(
        self, filter: str, cxt: dict[str, t.Any] | None = None, limit: int | None = None, cursor: str | None = None
    ) -> AuthzLookupResult: ...


class Repository[T: EntityState](t.Protocol):

    async def save(self, state: T) -> None: ...
    async def delete(self, criteria: Criteria) -> None: ...
    async def exists(self, criteria: Criteria) -> bool: ...
    async def get(self, criteria: Criteria) -> T | None: ...
    async def list(self, criteria: ListCriteria) -> ListPage[T]: ...


class DB[T = t.Any](t.Protocol):

    async def close(self) -> None: ...
    async def connect(self) -> None: ...
    def session(self) -> c.AsyncGenerator[T]: ...
    def transaction(self) -> c.AsyncGenerator[T]: ...



import re
import os
import inspect
import functools
import typing as t
from ulid import ulid
from json import loads
import collections.abc as c
from zoneinfo import ZoneInfo
from datetime import datetime, timezone
from contextlib import asynccontextmanager


__all__ = [
    "to_tz",
    "new_ulid",
    "safe_call",
    "from_json",
    "serialize",
    "impartial",
    "sanitize_dict",
    "to_camel_case",
    "to_snake_case",
    "get_signature",
    "unwrapped_call",
    "get_config_var",
    "to_context_manager",
]


new_ulid: t.Callable[[], str] = lambda: ulid().lower()


def from_json(val: t.Any) -> t.Any:
    try:
        return loads(val)
    except:
        return None


def sanitize_dict(obj: dict[str, t.Any]) -> dict[str, t.Any]:
    return {
        k: sanitize_dict(t.cast(dict[str, t.Any], v)) if isinstance(v, dict) else v
        for k, v in obj.items()
        if v is not None
    }


def to_snake_case(val: t.Any) -> t.Any:
    if isinstance(val, dict):
        return {to_snake_case(k): to_snake_case(v) for k, v in val.items()}
    if isinstance(val, list):
        return [to_snake_case(i) for i in val]
    if isinstance(val, str):
        # Insert underscores before capitals and lowercase everything
        s = re.sub(r"(.)([A-Z][a-z]+)", r"\1_\2", val)
        return re.sub(r"([a-z0-9])([A-Z])", r"\1_\2", s).lower()
    return val


def to_camel_case(val: t.Any) -> t.Any:
    if isinstance(val, dict):
        return {to_camel_case(k): to_camel_case(v) for k, v in val.items()}
    if isinstance(val, list):
        return [to_camel_case(i) for i in val]
    if isinstance(val, str):
        components = val.split("_")
        return components[0] + "".join(x.title() for x in components[1:])
    return val


def serialize(data: t.Any) -> t.Any:
    if hasattr(data, "serialize") and callable(getattr(data, "serialize")):
        return data.serialize()
    if isinstance(data, (list, tuple)):
        return [serialize(x) for x in data]
    if isinstance(data, dict):
        res_dict = {}
        for key, data in data.items():
            res_dict[key] = serialize(data)
        return res_dict
    return data


def iso_to_timestamp(date_string: str, format: str = "%Y-%m-%dT%H:%M:%SZ"):
    dt = datetime.strptime(date_string, format)
    dt = dt.replace(tzinfo=timezone.utc)
    return int(dt.timestamp() * 1000)


def timestamp_to_iso(timestamp: int, format: str = "%Y-%m-%dT%H:%M:%SZ") -> str:
    dt = datetime.fromtimestamp(timestamp / 1000, tz=timezone.utc)
    return dt.strftime(format)


def find_item[T](match_with: t.Callable[[T], bool], items: list[T]) -> T | None:
    found_item: T | None = None
    for item in items:
        if match_with(item):
            found_item = item
            break
    return found_item


def to_tz(tz_str: str) -> ZoneInfo:
    try:
        return ZoneInfo(key=tz_str)
    except:
        return ZoneInfo("UTC")


def names_from_email(email: str) -> tuple[str, str]:
    names = email.split("@")[0].split(".")
    if len(names) == 0:
        return "", ""
    if len(names) == 1:
        return names[0], ""
    return " ".join(names[:-1]), names[-1]


def safe_call[T](func: t.Callable[..., T], *args: t.Any, **kwargs: t.Any) -> T | None:
    try:
        return func(*args, **kwargs)
    except:
        return None


@asynccontextmanager
async def to_context_manager(gen: c.AsyncGenerator[t.Any, t.Any]):
    try:
        value = await gen.__anext__()
        yield value
    finally:
        await gen.aclose()


def get_signature(fn: t.Callable[..., t.Any]) -> list[inspect.Parameter]:
    unwrapped = unwrapped_call(fn)
    if unwrapped is None:
        return []
    return list(inspect.signature(unwrapped).parameters.values())


def unwrapped_call(call: t.Callable[..., t.Any] | None) -> t.Callable[..., t.Any] | None:
    if call is None:
        return call
    unwrapped = inspect.unwrap(impartial(call))
    return unwrapped


def impartial(func: t.Callable[..., t.Any]) -> t.Callable[..., t.Any]:
    while isinstance(func, functools.partial):
        func = func.func
    return func


def get_config_var(keys: list[str], default: str | None = None) -> t.Callable[[], str | None]:

    def get_var() -> str | None:
        for key in keys:
            if "/" in key:
                with open(key, "r") as f:
                    return f.read().strip()
            value = os.getenv(key)
            if value:
                return value
        return default

    return get_var



import typing as t
from attrs import define, field
from asyncio import Queue, Event, create_task, wait

from skoll.domain import Message
from skoll.result import Result, fail
from skoll.application import Subscriber, MissingSubscriber, Mediator, Service, RawMessage

from .utils import run_callback, is_subscribed

__all__ = ["BasicMediator"]


@define(kw_only=True, slots=True)
class BasicMediator(Mediator):
    """Basic in-memory mediator implementation."""

    queue: Queue[Message | RawMessage] = field(factory=Queue)
    stop_signal: Event = field(factory=Event)
    subscribers: list[Subscriber] = field(factory=list)
    handlers: dict[str, Subscriber] = field(factory=dict)

    @t.override
    async def publish(self, msg: Message | RawMessage) -> None:
        self.queue.put_nowait(msg)

    @t.override
    async def disconnect(self):
        self.stop_signal.set()

    @t.override
    async def request(self, msg: Message | RawMessage) -> Result[t.Any]:
        topic = msg.topic if isinstance(msg, Message) else msg["topic"]
        subscriber = self.handlers.get(topic)
        if subscriber is not None:
            return await run_callback(subscriber, msg)
        return fail(err=MissingSubscriber(debug={"subject": topic}))

    @t.override
    def register(self, *services: Service) -> None:
        for service in services:
            for subscriber in service.subscribers:
                if subscriber.with_reply:
                    self.handlers[subscriber.topic] = subscriber
                else:
                    self.subscribers.append(subscriber)

    @t.override
    async def connect(self):
        while not self.stop_signal.is_set() or not self.queue.empty():
            try:
                msg = await self.queue.get()
                tasks = [
                    create_task(run_callback(subscriber, msg))
                    for subscriber in self.subscribers
                    if is_subscribed(subscriber, msg)
                ]
                await wait(tasks)
            except Exception as exc:
                print(exc)
